/*
** Module   : DatabaseTriggerManager.java
** Abstract : Database trigger manager.
**
** Copyright (c) 2009-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ---------------------------- Description ----------------------------
** 001 SVL 20091019   @44152 Created initial version. Supports assign and 
**                           delete schema triggers.
** 002 CA  20100527   @44879 Added support for CREATE and "no record is 
**                           available triggers". The CREATE, DELETE and 
**                           ASSIGN triggers will execute on undo too, but the
**                           passed DMO will not be the buffer proxy; instead,
**                           it will be the actual record instance.
** 003 SVL 20130213          Added (de)registerDatabaseTrigger and disable*Triggers functions.
** 004 OM  20131107          Rewritten entire class.
*/

package com.goldencode.p2j.persist.trigger;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.util.*;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.*;

/**
 * This class handles the management of trigger system for a database. It maintains a stack of
 * scopes (a layer of triggers for each subroutine block).
 * One instance of this class is associated with a single, physical database, per user context.
 * <p>
 * In a context, there can be multiple managers, one for each used database. Each of them is
 * notified when the execution point enters and leaves subroutine blocks.
 */
public class DatabaseTriggerManager implements Scopeable
{
   /** Context local map of instances of this class, mapped by theirs database name . */
   private static final ContextLocal<DatabaseTriggerManager> context =
         new ContextLocal<DatabaseTriggerManager>()
         {
            protected DatabaseTriggerManager initialValue()
            {
               return new DatabaseTriggerManager();
            }
         };

   /**
    * The 'global' layer of schema triggers.
    */
   private static final TriggerBlockLayer schemaTriggers = new TriggerBlockLayer();

   /**
    * The stack of all layer of triggers handled by this manager.
    */
   private Deque<TriggerBlockLayer> stack = new ArrayDeque<>();

   /**
    * Private constructor. An instance of this class can be obtained only using get() method.
    * Registers this object with TransactionManager as a Scopeable instance.
    */
   private DatabaseTriggerManager()
   {
      // register me with TransactionManager to receive enter/leave bloc events
      TransactionManager.registerScopeable(this);
   }

   /**
    * Recursively look for DatabaseTrigger class in the <code>dir</code> directory and all
    * its subdirs.
    *
    * @param   dir
    *          The directory to look for classes.
    * @param   packageName
    *          The package that is associated with dir directory.
    * @param   recursive
    *          true for recursive search.
    */
   private static void getTriggerClasses(File dir,
                                         String packageName,
                                         List<Class<? extends DatabaseTrigger>> classes,
                                         boolean recursive)
   {
      File[] files = dir.listFiles();
      if (files == null)
      {
         // could not list files from dir
         return;
      }
      for (File file : files) {
         String fn = file.getName();
         if (file.isDirectory())
         {
            if (recursive)
            {
               // call recursively for sub-dir/package
               getTriggerClasses(file, packageName + "." + fn, classes, true);
            }
         }
         else if (fn.endsWith(".class"))
         {
            try
            {
               Class<?> ufo = Class.forName(packageName + '.' + fn.substring(0, fn.length() - 6));
               if (DatabaseTrigger.class.isAssignableFrom(ufo) &&
                     ufo.isAnnotationPresent(SchemaTrigger.class))
               {
                  // if the found class has the required interface add to return list
                  classes.add((Class<? extends DatabaseTrigger>) ufo);
               }
            }
            catch (ClassNotFoundException e)
            {
               // failed loading a class, perhaps wrong package ?
            }
         }
      }
   }

   /**
    * Initialization of <code>DatabaseTriggerManager</code> static internal data.
    * The most important step is checking the java path for classes that were generated with
    * <code>SchemaTrigger</code> annotation and prepare the schema trigger map.
    * The method scans all classes accessible from the context class loader which belong to
    * the given package and subpackages (if recursive is true).
    *
    * @param   triggersLocation
    *          the package from where the triggers are loaded (as declared in package statement)
    * @param   recursive
    *          search for triggers recursively
    */
   public static void initialize(String triggersLocation, boolean recursive)
   {
      List<Class<? extends DatabaseTrigger>> schemaTriggersList = new ArrayList<>();

      try
      {
         Enumeration<URL> resources = Thread.currentThread().getContextClassLoader()
                           .getResources(triggersLocation.replace('.', '/'));
         List<String> dirs = new ArrayList<>();
         while (resources.hasMoreElements())
         {
            dirs.add(resources.nextElement().getFile());
         }
         for (String directory : dirs)
         {
            getTriggerClasses(new File(directory),
                              triggersLocation,
                              schemaTriggersList,
                              recursive);
         }
      }
      catch (IOException e)
      {
         // could not obtain URL resource from context class loader, nothing to initialize
         return;
      }

      for (Class<? extends DatabaseTrigger> trigger : schemaTriggersList)
      {
         DatabaseEventType det = null;
         Class<? extends Buffer> bufferCls = null;
         String assignProperty = null;
         boolean override = false;
         boolean hasOldBuffer = false;

         if (trigger.isAnnotationPresent(SchemaTrigger.class))
         {
            // normally all triggers will have this annotation
            SchemaTrigger annotation = trigger.getAnnotation(SchemaTrigger.class);

            bufferCls = annotation.buffer();
            det = annotation.event();
            assignProperty = annotation.field();
            hasOldBuffer = annotation.hasOldBuffer();
            override = annotation.overrideable();
         }

         schemaTriggers.setTrigger(trigger,
                                   det,
                                   bufferCls,
                                   assignProperty,
                                   null, /* no container needed */
                                   override,
                                   hasOldBuffer);
      }

   }


   /**
    * Obtain the context-local DatabaseTriggerManager object.
    *
    * @return  the DatabaseTriggerManager for current context.
    */
   public static DatabaseTriggerManager get()
   {
      return context.get();
   }

   /**
    * Register a session database trigger within the current scope.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    * @param    triggerCls
    *           The class of the trigger.
    * @param    container
    *           The containing object (all triggers are implemented as non-static inner classes,
    *           so they must have a reference to the containing class).
    */
   public static void registerDatabaseTrigger(DatabaseEventType event,
                                              Class<? extends Buffer> bufferCls,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container)
   {
      registerDatabaseTrigger(event, bufferCls, triggerCls, container, false, false);
   }

   /**
    * Register a session database trigger within the current scope.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    * @param    triggerCls
    *           The class of the trigger.
    * @param    container
    *           The containing object (all triggers are implemented as non-static inner classes,
    *           so they must have a reference to the containing class).
    * @param    override
    *           <code>true</code> if OVERRIDE keyword is specified.
    */
   public static void registerDatabaseTrigger(DatabaseEventType event,
                                              Class<? extends Buffer> bufferCls,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container,
                                              boolean override)
   {
      registerDatabaseTrigger(event, bufferCls, triggerCls, container, override, false);
   }

   /**
    * Register a session database trigger within the current scope.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    * @param    triggerCls
    *           The class of the trigger.
    * @param    container
    *           The containing object (all triggers are implemented as non-static inner classes,
    *           so they must have a reference to the containing class).
    * @param    override
    *           <code>true</code> if OVERRIDE keyword is specified.
    */
   public static void registerDatabaseTrigger(DatabaseEventType event,
                                              Class<? extends Buffer> bufferCls,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container,
                                              boolean override,
                                              boolean hasOldBuffer)
   {
      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         // sanity check failed
         return;
      }

      // check if this event is overrideable
      if (override && !isSchemaTriggerOverrideable(event, bufferCls))
      {
         return;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.setTrigger(triggerCls, event, bufferCls, null, container, override, hasOldBuffer);
   }

   public static void registerDatabaseTrigger(DatabaseEventType event, // always ASSIGN
                                              Class<? extends Buffer> bufferCls,
                                              String property,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container)
   {
      registerDatabaseTrigger(event, bufferCls, property, triggerCls, container, false, false);
   }

   public static void registerDatabaseTrigger(DatabaseEventType event, // always ASSIGN
                                              Class<? extends Buffer> bufferCls,
                                              String property,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container,
                                              boolean override)
   {
      registerDatabaseTrigger(event, bufferCls, property, triggerCls, container, override, false);
   }

   public static void registerDatabaseTrigger(DatabaseEventType event, // always ASSIGN
                                              Class<? extends Buffer> bufferCls,
                                              String property,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container,
                                              boolean override,
                                              boolean hasOldValue)
   {
      // TODO: drop event from the list of parameters for this kind of events ?
      if (event != DatabaseEventType.ASSIGN)
      {
         throw new IllegalArgumentException("Use this form of register only for ASSIGN triggers.");
      }
      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         // sanity check failed
         return;
      }

      // check if this event is overrideable
      if (override && !isSchemaTriggerOverrideable(DatabaseEventType.ASSIGN, bufferCls))
      {
         return;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.setTrigger(triggerCls, event, bufferCls, property, container, override, hasOldValue);
   }

   /**
    * Deregister a session database trigger for the given event type and buffer class.
    * This is only called when a ON <event> REVERT.
    *
    * @param   event
    *          Database event type.
    * @param   bufferCls
    *          The class of the buffer to which the event is applied.
    * @param   property
    *          The property that is deregistered from ASSIGN trigger.
    */
   public static void deregisterDatabaseTrigger(DatabaseEventType event,
                                                Class<? extends Buffer> bufferCls,
                                                String property)
   {
      // TODO: drop event from the list of parameters for this kind of events ?

      if (event != DatabaseEventType.ASSIGN)
      {
         throw new IllegalArgumentException("Use this form of register only for ASSIGN triggers.");
      }
      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         // sanity check failed
         return;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.removeTrigger(event, bufferCls, property);
   }

   /**
    * Deregister a session database trigger for the given event type and buffer class.
    * This is only called when a ON <event> REVERT.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    */
   public static void deregisterDatabaseTrigger(DatabaseEventType event,
                                                Class<? extends Buffer> bufferCls)
   {
      if (event == DatabaseEventType.ASSIGN)
      {
         throw new IllegalArgumentException("Not executing ASSIGN with this method.");
      }

      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         // sanity check failed
         return;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.removeTrigger(event, bufferCls, null);
   }

   /**
    * Disables dump triggers (FIND queries) for the current block.
    *
    * @param   bufferCls
    *          The buffer for which to disable dump triggers.
    *
    * @return  <code>true</code> if the filter was successfully applied. If <code>false</code>
    *          is returned then the internal stack is broken because of bad block management.
    */
   public static boolean disableDumpTriggers(Class<? extends Buffer> bufferCls)
   {
      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         return false;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.disableTriggers(bufferCls, true, false, false);
      return true;
   }

   /**
    * Disables load triggers (CREATE, ASSIGN, WRITE queries) for the current block.
    *
    * @param   bufferCls
    *          The buffer for which to disable looad triggers.
    *
    * @return  <code>true</code> if the filter was successfully applied. If <code>false</code>
    *          is returned then the internal stack is broken because of bad block management.
    */
   public static boolean disableLoadTriggers(Class<? extends Buffer> bufferCls, boolean allowReplication)
   {
      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         return false;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.disableTriggers(bufferCls, false, true, !allowReplication);
      return true;
   }

   /**
    * Executes a trigger. (Except for ASSIGN triggers.)
    *
    * @param   det
    *          The trigger type to be executed.
    * @param   bufferCls
    *          The buffer class for the dmo that fired the trigger.
    * @param   newVal
    *          The buffer that fired the trigger.
    * @param   oldDmoValue
    *          When the (WRITE) trigger declares an 'old' buffer value, this is passed to
    *          trigger procedure as r/o buffer.
    */
   public static void trigger(DatabaseEventType det,
                              Class<? extends Buffer> bufferCls,
                              Buffer newVal,
                              Persistable oldDmoValue)
   {
      // TODO: Persistence.load():2029: Object o = query.uniqueResult();
      // TODO: can trigger an uncatched WRITE database event

      DatabaseTriggerManager manager = get();
      if (!manager.isTriggerEnabled(det, bufferCls))
      {
         // triggers are disabled for this kind of event / table
         return;
      }

      // look for the triggers for this event / bufferCls
      TriggerData sessionTrigger = manager.peekTrigger(det, bufferCls, null);
      TriggerData schemaTrigger = schemaTriggers.getTrigger(det, bufferCls, null);
      boolean disableSchemaTrigger = (sessionTrigger != null) && sessionTrigger.isOverride();

      // each trigger invocation is done using a different instance of the OLD, r/o buffer
      // it will be constructed in place when needed.
      boolean sessOldBuffer =
            oldDmoValue != null && sessionTrigger != null && sessionTrigger.hasOldBuffer();
      boolean schOldBuffer =
            oldDmoValue != null && schemaTrigger != null && schemaTrigger.hasOldBuffer();

      // except for the FIND case, the session trigger executes first (if it ever does)
      if (sessionTrigger != null && det != DatabaseEventType.FIND)
      {
         Buffer oldValue = sessOldBuffer ? createWriteOldBuffer(bufferCls, oldDmoValue) : null;
         fireTrigger(det, sessionTrigger, newVal, oldValue);
      }

      if (schemaTrigger != null && !disableSchemaTrigger)
      {
         Buffer oldValue = schOldBuffer ? createWriteOldBuffer(bufferCls, oldDmoValue) : null;
         fireTrigger(det, schemaTrigger, newVal, oldValue);
      }

      // in the FIND case, the session trigger executes second (if it ever does)
      if (sessionTrigger != null && det == DatabaseEventType.FIND)
      {
         Buffer oldValue = sessOldBuffer ? createWriteOldBuffer(bufferCls, oldDmoValue) : null;
         fireTrigger(det, sessionTrigger, newVal, oldValue);
      }
   }

   /**
    * Executes an ASSIGN trigger.
    *
    * @param   bufferCls
    *          The buffer class for the dmo that fired the trigger.
    * @param   newVal
    *          The buffer that contains the field that fired the trigger.
    * @param   oldValue
    *          When the ASSIGN trigger declares an 'old' value, this is passed to trigger
    *          procedure. The procedure code is allowed to modify this value without generating
    *          any error, however, they are lost as only a duplicate is passed on.
    */
   public static void triggerAssign(Class<? extends Buffer> bufferCls,
                                    Buffer newVal,
                                    String property,
                                    BaseDataType oldValue)
   {
      DatabaseTriggerManager manager = get();
      if (!manager.isTriggerEnabled(DatabaseEventType.ASSIGN, bufferCls))
      {
         // triggers are disabled for this kind of event / buffer
         return;
      }

      //look for the trigger
      TriggerData sessionTrigger = manager.peekTrigger(DatabaseEventType.ASSIGN,
                                                       bufferCls,
                                                       property);
      TriggerData schemaTrigger = schemaTriggers.getTrigger(DatabaseEventType.ASSIGN,
                                                            bufferCls,
                                                            null);
      boolean disableSchemaTrigger = (sessionTrigger != null) && sessionTrigger.isOverride();

      //execute session trigger first
      if (sessionTrigger != null)
      {
         fireAssignTrigger(sessionTrigger, newVal, oldValue.duplicate());
      }

      // the schema trigger executes second (if it ever does)
      if (!disableSchemaTrigger && schemaTrigger != null)
      {
         fireAssignTrigger(schemaTrigger, newVal, oldValue.duplicate());
         // if trigger object failed an exception have already been printed to console
      }
   }

   /**
    * Provides a notification that a new scope is about to be entered. The manager only acts on
    * top-level scope. Each time a new scope is entered a fresh new layer of triggers is added
    * to internal stack.
    */
   @Override
   public void scopeStart()
   {
      // triggers get notifications at all top level blocks
      if (TransactionManager.isTopLevelBlock())
      {
         stack.addFirst(new TriggerBlockLayer()); // push();
      }
   }

   /**
    * Provides a notification that a scope is about to be exited. The manager only acts on
    * top-level scope. Each time a scope is exited a the last layer of triggers is removed from
    * internal stack.
    */
   @Override
   public void scopeFinished()
   {
      // triggers get notifications at all top level blocks
      if (TransactionManager.isTopLevelBlock())
      {
         stack.removeFirst(); // pop();
      }
   }

   /**
    * Tests if a schema trigger is overrideable. In case the trigger is declared as overideable,
    * <code>true</code> is returned. Otherwise, 3223 code is <code>recordOrThrowError</code>.
    * If no error is acting <code>false</code> will be returned.
    *
    * @param   event
    *          The database event of the trigger to be tested.
    * @param   bufferCls
    *          The class buf for the table the trigger is defined.
    *
    * @return  <code>true</code> if respective trigger can be overrided by a session trigger.
    */
   private static boolean isSchemaTriggerOverrideable(DatabaseEventType event,
                                                      Class<? extends Buffer> bufferCls)
   {
      // check if this event is overrideable
      TriggerData schemaTrigger = schemaTriggers.getTrigger(event, bufferCls, null);
      if (schemaTrigger != null && !schemaTrigger.isOverride())
      {
         // detect table name
         String tableName = "?";
         for (Class<?> next : bufferCls.getInterfaces())
         {
            if (DataModelObject.class.isAssignableFrom(next))
            {
               tableName = next.getSimpleName();
               break;
            }
         }

         String msg = String.format("Attempt to OVERRIDE a schema %s trigger for %s",
                                    event,
                                    tableName);
         // this is a non overrideable schema trigger:
         ErrorManager.recordOrThrowError(3223, msg);
         return false;
      }
      return true;
   }


   /**
    * Create a trigger object instance based on information from a <code>TriggerData</code>
    * container.
    *
    * @param   triggerData
    *          The container that holds information about the trigger to be built.
    *
    * @return  An instance of the trigger as configured in <code>triggerData</code>.
    */
   private static DatabaseTrigger instantiateTrigger(TriggerData triggerData) {
      try
      {
         if (triggerData.getTriggerContainer() == null)
         {
            // if the container is null then this is (probably) a schema trigger
            // being a top-level public class, it does not need a container for constructing
            return triggerData.getTriggerClass().newInstance();
         }

         Class[] ctorParmlist  = new Class[] { triggerData.getTriggerContainer().getClass() };
         Constructor<? extends DatabaseTrigger> ctor =
               triggerData.getTriggerClass().getConstructor(ctorParmlist);
         return ctor.newInstance(triggerData.getTriggerContainer());
      }
      catch (InstantiationException | IllegalAccessException |
            NoSuchMethodException | InvocationTargetException e)
      {
         e.printStackTrace();
         // could not instantiate the trigger class
         return null;
      }
   }

   /**
    * Extract only the DMO schema-specific interface (i.e., exclude the Buffer interface)
    *
    * @param   bufferCls
    *          Interface which defines DMO's public API plus all {@link Buffer} methods.
    * @param   oldDmoValue
    *          The read-only value.
    *
    * @return  A temporary read-only buffer that can be used as the OLD 'parameter' in a
    *          WRITE trigger.
    */
   private static Buffer createWriteOldBuffer(Class<? extends Buffer> bufferCls,
                                              Persistable oldDmoValue)
   {
      Class<?> dmoIface = null;
      Class<?>[] superIfaces = bufferCls.getInterfaces();
      for (Class<?> next : superIfaces)
      {
         if (DataModelObject.class.isAssignableFrom(next))
         {
            dmoIface = next;
            break;
         }
      }

      if (dmoIface == null)
      {
         // TODO: throw something here ?
         return null;
      }

      Buffer oldValue = RecordBuffer.defineReadOnly(bufferCls,
                                                    DatabaseManager.getSchemaByInterface(dmoIface),
                                                    oldDmoValue);
      oldValue.openReadOnlyScope();
      // TODO: already open, open here or after entering the trigger block ?
      return oldValue;
   }

   /**
    * Instantiates an object of the ASSIGN trigger and fires it.
    *
    * @param   triggerData
    *          Information about how the trigger will be built.
    * @param   newVal
    *          The buffer that contains the field that fired the trigger.
    * @param   oldValue
    *          The old value of the field. May be null if the trigger did not declared it.
    */
   private static void fireAssignTrigger(TriggerData triggerData,
                                         Buffer newVal,
                                         BaseDataType oldValue)
   {
      final DatabaseTrigger trigger = instantiateTrigger(triggerData);
      if (trigger != null)
      {
         if (triggerData.hasOldBuffer()) {
            trigger.assign(newVal, oldValue);
         }
         else
         {
            trigger.assign(newVal);
         }
      }
      // if trigger object failed an exception have already been printed to console
   }

   /**
    * Instantiates a trigger object associated with the specified database event type and
    * fires it.
    *
    * @param   det
    *          The database event type to be fired.
    * @param   triggerData
    *          Information about how the trigger will be built.
    * @param   buffer
    *          The buffer that fired the trigger.
    * @param   oldValue
    *          The old value of the buffer. May be null if the trigger did not declared it.
    */
   private static void fireTrigger(DatabaseEventType det,
                                              TriggerData triggerData,
                                              Buffer buffer,
                                              Buffer oldValue)
   {
      final DatabaseTrigger trigger = instantiateTrigger(triggerData);
      if (trigger == null)
      {
         // if trigger object failed an exception have already been printed to console
         return;
      }

      switch (det)
      {
         case ASSIGN:
            throw new IllegalArgumentException("Not executing ASSIGN with this method.");
            // break;
         case CREATE:
            trigger.create(buffer);
            break;
         case DELETE:
            trigger.delete(buffer);
            break;
         case FIND:
            trigger.find(buffer);
            break;
         case WRITE:
            if (triggerData.hasOldBuffer() /* && oldValue != null ? */)
            {
               trigger.write(buffer, oldValue);
            }
            else
            {
               trigger.write(buffer);
            }
            break;
         case REPLICATION_CREATE:
            trigger.replicationCreate(buffer);
            break;
         case REPLICATION_DELETE:
            trigger.replicationDelete(buffer);
            break;
         case REPLICATION_WRITE:
            if (triggerData.hasOldBuffer() /* && oldValue != null ? */)
            {
               trigger.replicationWrite(buffer, oldValue);
            }
            else
            {
               trigger.replicationWrite(buffer);
            }
            break;
      }
   }

   public static void handleError(DatabaseEventType det,
                                  Class<? extends Buffer> bufferClass,
                                  String bufferName,
                                  String fieldName)
   {
      switch (det)
      {
         case CREATE:
            ErrorManager.showErrorAndAbend(3168,
                  "Attempt to CREATE a " + bufferClass +
                  "record in CREATE-TRIGGER " + ProcedureManager.lastProcedure() +
                  "while a trigger is executing");
            break;
         case DELETE:
            ErrorManager.showErrorAndAbend(3169,
                  "Attempt to DELETE a " + bufferClass +
                  "record in DELETE-TRIGGER " + ProcedureManager.lastProcedure() +
                  "while a trigger is executing");
            break;
         case FIND:
            ErrorManager.displayError(2873,
                  "Cannot change cursor of buffer " + bufferName +
                  "when executing a trigger");
            break;
         case ASSIGN:
            ErrorManager.displayError(1003,
               "Cycle in procedure calls detected in 'ASSIGN-TRIGGER " +
               ProcedureManager.lastProcedure() + "' while executing ASSIGN trigger for " +
               bufferName + "." + fieldName);
            break;
      }
   }

   /**
    * Looks into the stack of layers and check if a trigger is enabled for a database event.
    *
    * @param   det
    *          The database event.
    * @param   bufferCls
    *          The buffer.
    * @return  true if the trigger is enabled.
    */
   private boolean isTriggerEnabled(DatabaseEventType det, Class<? extends Buffer> bufferCls)
   {
      for (TriggerBlockLayer tbl : stack)
      {
         if (tbl.isTriggerDisabled(det, bufferCls))
         {
            return false;
         }
      }
      return true;
   }

   /**
    * Implementation of the trigger search algorithm in the stack of triggers. Each layer of the
    * stack is inspected for <code>det</code> type of trigger associated with <code>buff</code>.
    * If found it is returned, otherwise the search continue with the next level from the stack.
    * If no such trigger can be found, <code>null</code> is returned.
    * This method takes into account if respective trigger event was disabled and, in this case,
    * a special singleton object is returned.
    *
    * @param   det
    *          The database trigger type.
    * @param   bufferCls
    *          The buffer that raised the trigger event.
    *
    * @return  The top-level instance of the trigger data of the <code>det</code> type associated
    *          with <code>buff</code>.
    */
   private TriggerData peekTrigger(DatabaseEventType det,
                                   Class<? extends Buffer> bufferCls,
                                   String property)
   {
      for (TriggerBlockLayer blockLayer : stack) // the proper order is guaranteed by deque.iterator() method
      {
         final TriggerData triggerData = blockLayer.getTrigger(det, bufferCls, property);
         if (triggerData == null)
         {
            // no trigger found at this level, search deeper
            continue;
         }

         // if disabledTrigger was returned either a trigger was found or not, it does not
         // matter as long as nothing will be executed (disabledTrigger methods are empty and
         // also isOverride() return true so schema trigger is also disabled).
         // if a normal trigger is found it will be executed along with schema trigger
         return triggerData;
      }

      // search failed, no trigger of type det was defined for buff
      return null;
   }
}

/**
 * This class holds all triggers defined ar a certain level in the stack of triggers along with
 * information whether some groups of triggers are disabled. The class also provides methods
 * for installing and removing triggers into/from this layer and for obtaining the active
 * <code>DatabaseTrigger</code> for current layer.
 */
class TriggerBlockLayer
{
   /** A FIND <code>TriggerData</code> can be defined for each table on this layer.*/
   private Map<Class<? extends Buffer>, TriggerData> findTriggers = null;

   /** A CREATE <code>TriggerData</code> can be defined for each table on this layer.*/
   private Map<Class<? extends Buffer>, TriggerData> createTriggers = null;

   /** A WRITE <code>TriggerData</code> can be defined for each table on this layer.*/
   private Map<Class<? extends Buffer>, TriggerData> writeTriggers = null;

   /**
    * A DELETE <code>TriggerData</code> can be defined for each table on this layer.
    * TODO: investigate disable delete triggers.
    */
   private Map<Class<? extends Buffer>, TriggerData> deleteTriggers = null;

   /**
    * For each field (property) of each table an ASSIGN <code>TriggerData</code> can be on this
    * layer. The mapping order is rather uncommon but this help re-using the code from lookup
    * of other triggers by applying first the mapping from field name.
    */
   private Map<String, Map<Class<? extends Buffer>, TriggerData>> assignTriggers = null;

   /**
    * A REPLICATION-CREATE <code>TriggerData</code> can be defined for each table on this layer.
    * Not implemented at this time.
    */
   private Map<Class<? extends Buffer>, TriggerData> repCreateTriggers = null;

   /**
    * A REPLICATION-WRITE <code>TriggerData</code> can be defined for each table on this layer.
    * Not implemented at this time.
    */
   private Map<Class<? extends Buffer>, TriggerData> repWriteTriggers = null;

   /**
    * A REPLICATION-DELETE <code>TriggerData</code> can be defined for each table on this layer.
    * Not implemented at this time.
    */
   private Map<Class<? extends Buffer>, TriggerData> repDeleteTriggers = null;

   /** This is the set of tables that disabled their DUMP triggers (FIND). */
   private Set<Class<? extends Buffer>> disabledDumpTriggers = new HashSet<>();

   /** This is the set of tables that disabled their LOAD triggers (CREATE, WRITE, ASSIGN, DELETE?). */
   private Set<Class<? extends Buffer>> disabledLoadTriggers = new HashSet<>();

   /** This is the set of tables that disabled their REPLICATION triggers. */
   private Set<Class<? extends Buffer>> disabledReplicationTriggers = new HashSet<>();

   /**
    * Constructor with minimum implementation, all members will use lazy initialization for
    * two reasons:
    *    * to not allocate unneeded memory
    *    * quick creation of the object.
    */
   TriggerBlockLayer()
   {
   }

   /**
    * Obtain the container that hold information for instantiating a trigger object.
    *
    * @param   det
    *          The database event to which the trigger responds to.
    * @param   bufferCls
    *          The class that defines the table for which the trigger applies to.
    * @param   property
    *          Optionally, for ASSIGN triggers the name of the property (field) for which the
    *          trigger was declared. Otherwise it is ignored.
    *
    * @return  The container that holds information of how to create an instance of the specified
    *          trigger and how to apply it of null if such session trigger has not been declared
    *          within the current layer.
    */
   TriggerData getTrigger(DatabaseEventType det,
                          Class<? extends Buffer> bufferCls,
                          String property)
   {
      // look for the maps of buffer-to-triggers associated with det event
      Map<Class<? extends Buffer>, TriggerData> triggerMap =
            getBufferDatabaseTriggerMap(det, property, false);
      if (triggerMap == null)
      {
         // no trigger of this kind was declared at this block level
         return null;
      }

      // if a trigger was set for this buffer return it, otherwise return null
      return triggerMap.get(bufferCls);
   }

   /**
    * Sets the current trigger for a <code>det</code> and a <code>buffer</code>. If another
    * trigger was set for the same event and buffer, it will be dropped.
    * Because this class uses lazy initialization the map of trigger is created automatically
    * if there it was not already there for this trigger event.
    *
    * @param   triggerCls
    *          The trigger to be set.
    * @param   det
    *          The trigger event.
    * @param   bufferCls
    *          The buffer for which the trigger is set.
    * @param   assignProperty
    *          When setting an ASSIGN trigger the name of the field is needed, otherwise ignored.
    * @param   container
    *          The container object of the trigger.
    * @param   override
    *          If <code>true</code>, the schema trigger is blocked, otherwise, schema trigger
    *          will be executed just after the session trigger.
    * @param   hasOldBuffer
    *          <code>true</code> if the callback of the trigger has reference to old buffer
    */
   void setTrigger(Class<? extends DatabaseTrigger> triggerCls,
                   DatabaseEventType det,
                   Class<? extends Buffer> bufferCls,
                   String assignProperty,
                   Object container,
                   boolean override,
                   boolean hasOldBuffer)
   {
      // get the trigger map, creating it if needed
      Map<Class<? extends Buffer>, TriggerData> triggerMap =
            getBufferDatabaseTriggerMap(det, assignProperty, true);
      assert (triggerMap != null);

      TriggerData triggerData = new TriggerData(triggerCls, container, override, hasOldBuffer);

      // drop the old trigger, if any
      triggerMap.put(bufferCls, triggerData);
   }

   /**
    * Removes a trigger from the associated map.
    *
    * @param   det
    *          The event type of the trigger to be reverted
    * @param   bufferCls
    *          The buffer for which the trigger is removed.
    * @param   property
    *          The property that is deregistered from ASSIGN trigger.
    */
   void removeTrigger(DatabaseEventType det, Class<? extends Buffer> bufferCls, String property)
   {
      // get the trigger map, possible null
      Map<Class<? extends Buffer>, TriggerData> triggerMap =
            getBufferDatabaseTriggerMap(det, property, false);

      if (triggerMap != null)
      {
         // drop the old trigger
         triggerMap.remove(bufferCls);

         // if triggerMap.isEmpty() does it worth it to set to null ?
      }

      // TODO: if the triggerMap is null or it does not contain a trigger for specified buffer
      // then there might be an incorrect pairing of ON event / ON event REVERT in user code.
      // should display some kind of message in this case ?
   }

   /**
    * Disable a set of trigger events. The disabled set of triggers will remain disabled until
    * the end of the scope of the current layer (the end of the current subroutine block).
    * Individual sets of triggers are disabled. For the set of trigger that are not explicitly
    * disabled the old status remain unchanged.
    *
    * @param   dump
    *          Disable the FIND triggers.
    * @param   load
    *          Disable the LOAD (ASSIGN, CREATE, WRITE) triggers.
    * @param   replication
    *          Disable replication triggers.
    */
   void disableTriggers(Class<? extends Buffer> bufferCls,
                        boolean dump,
                        boolean load,
                        boolean replication)
   {
      if (dump)
      {
         disabledDumpTriggers.add(bufferCls);
      }
      if (load)
      {
         disabledLoadTriggers.add(bufferCls);
      }
      if (replication)
      {
         disabledReplicationTriggers.add(bufferCls);
      }
   }

   /**
    * Lookup the buffer-to-trigger map using the DatabaseEventType. If the last parameter is
    * true and respective map is null, then it is created on-the-fly, otherwise the current
    * value is returned.
    *
    * @param   det
    *          The event type that is looked up.
    * @param   createIfNotExist
    *          Flag to create the specified map if the map has not already been created.
    *
    *
    * @return  the map that holds <code>det</code> buffer-to-events, possible <code>null</code>
    *          or a new empty map if this was requested.
    *
    */
   private Map<Class<? extends Buffer>, TriggerData> getBufferDatabaseTriggerMap(
         DatabaseEventType det, String assignField, boolean createIfNotExist)
   {
      switch (det)
      {
         case ASSIGN:
            if (assignTriggers == null)
            {
               if (createIfNotExist)
               {
                  assignTriggers = new HashMap<>();
               }
               else
               {
                  // no ASSIGN trigger defined for this table
                  return null;
               }
            }
            assert (assignTriggers != null); // evidently

            Map<Class<? extends Buffer>, TriggerData> triggersMap = assignTriggers.get(assignField);
            if (triggersMap == null && createIfNotExist)
            {
               triggersMap = new HashMap<>();
               assignTriggers.put(assignField, triggersMap);
            }
            return triggersMap;

         case CREATE:
            if (createTriggers == null && createIfNotExist)
            {
               createTriggers = new HashMap<>();
            }
            return createTriggers;

         case DELETE:
            if (deleteTriggers == null && createIfNotExist)
            {
               deleteTriggers = new HashMap<>();
            }
            return deleteTriggers;

         case FIND:
            if (findTriggers == null && createIfNotExist)
            {
               findTriggers = new HashMap<>();
            }
            return findTriggers;

         case WRITE:
            if (writeTriggers == null && createIfNotExist)
            {
               writeTriggers = new HashMap<>();
            }
            return writeTriggers;

         case REPLICATION_CREATE:
            if (repCreateTriggers == null && createIfNotExist)
            {
               repCreateTriggers = new HashMap<>();
            }
            return repCreateTriggers;

         case REPLICATION_DELETE:
            if (repDeleteTriggers == null && createIfNotExist)
            {
               repDeleteTriggers = new HashMap<>();
            }
            return repDeleteTriggers;

         case REPLICATION_WRITE:
            if (repWriteTriggers == null && createIfNotExist)
            {
               repWriteTriggers = new HashMap<>();
            }
            return repWriteTriggers;
      }

      // this should never been reached
      return null;
   }

   /**
    * Check if a trigger is disabled for a buffer.
    *
    * @param   det
    *          Database event type.
    * @param   bufferCls
    *          The buffer.
    *
    * @return  True if trigger was disabled.
    */
   public boolean isTriggerDisabled(DatabaseEventType det, Class<? extends Buffer> bufferCls)
   {
      if (det == DatabaseEventType.FIND)
      {
         return disabledDumpTriggers.contains(bufferCls);
      }

      if (det == DatabaseEventType.ASSIGN ||
          det == DatabaseEventType.CREATE ||
          det == DatabaseEventType.WRITE)
      {
         return disabledLoadTriggers.contains(bufferCls);
      }

      if (det == DatabaseEventType.REPLICATION_DELETE ||
          det == DatabaseEventType.REPLICATION_WRITE  ||
          det == DatabaseEventType.REPLICATION_CREATE)
      {
         return disabledReplicationTriggers.contains(bufferCls);
      }

      return true;
   }
}

/**
 * This is a container for all information needed for a trigger callback. This class is
 * not mutable, the values passed to constructor can be later obtained using the individual
 * getters.
 */
class TriggerData
{
   /** Trigger class that will be instantiated and executed. */
   private Class<? extends DatabaseTrigger> triggerClass;

   /** The container in which the trigger is declared. */
   private Object triggerContainer;

   /** Flag for disabling the schema trigger. */
   private boolean override;

   /** Flag that the trigger needs the old value of the buffer (WRITE only) */
   private boolean oldBuffer;

   /**
    * Builds an unmutable object with data needed for a trigger to instantiate/execute.
    *
    * @param   triggerClass
    *          The class of the trigger.
    * @param   triggerContainer
    *          The container of the trigger.
    * @param   override
    *          Allows to block the schema trigger or execute before it.
    * @param   oldBuffer
    *          Some triggers need a buffer with old value.
    */
   TriggerData(Class<? extends DatabaseTrigger> triggerClass,
               Object triggerContainer,
               boolean override,
               boolean oldBuffer)
   {
      this.triggerClass = triggerClass;
      this.triggerContainer = triggerContainer;
      this.override = override;
      this.oldBuffer = oldBuffer;
   }

   /**
    * Obtain the trigger class.
    *
    * @return  The trigger class.
    */
   Class<? extends DatabaseTrigger> getTriggerClass()
   {
      return triggerClass;
   }

   /**
    * Obtain the trigger container.
    *
    * @return  The container object of the trigger.
    */
   Object getTriggerContainer()
   {
      return triggerContainer;
   }

   /**
    * Check if this trigger replaces the schema trigger (if any) or if it executes before.
    *
    * @return  The override flag.
    */
   boolean isOverride()
   {
      return override;
   }

   /**
    * Check if this trigger additionally need the old value of the buffer.
    *
    * @return  true if this trigger additionally need the old value of the buffer.
    */
   boolean hasOldBuffer()
   {
      return oldBuffer;
   }
}
