/*
** Module   : DatabaseTriggerDelegate.java
** Abstract : Delegate object for invocation of database triggers.
**
** Copyright (c) 2009-2010, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 SVL 20091019   @44153 Created initial version. Supports assign and
**                           delete triggers.
** 002 CA  20100527   @44878 Added support for CREATE and "no record is 
**                           available triggers". The CREATE, DELETE and 
**                           ASSIGN triggers will execute on undo too, but the
**                           passed DMO will not be the buffer proxy; instead,
**                           it will be the actual record instance.
*/

package com.goldencode.p2j.persist.trigger_old;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

import java.lang.reflect.*;
import java.util.*;

/**
 * Provides a service for the fast invocation of database triggers for a
 * specific DMO interface.
 */
public class DatabaseTriggerDelegate
{
   /**
    * Set of definitions which represent create triggers for a specific DMO
    * interface.
    */
   private final Set<DatabaseTriggerDef> createTriggers;

   /**
    * Set of definitions which represent delete triggers for a specific DMO
    * interface.
    */
   private final Set<DatabaseTriggerDef> deleteTriggers;

   /**
    * Map of property names to sets of definitions which represent assign 
    * triggers for a given property (for a specific DMO interface).
    */
   private final Map<String, Set<DatabaseTriggerDef>> assignTriggers;

   /**
    * Set of definitions which represent "no record is available" triggers for 
    * a specific DMO interface.
    */
   private final Set<DatabaseTriggerDef> noRecordAvailTriggers;

   /**
    * Creates the delegate for a specific DMO interface.
    *
    * @param createTriggers
    *        Set of definitions which represent create triggers for a specific 
    *        DMO interface. May be <code>null</code>.
    * @param deleteTriggers
    *        Set of definitions which represent delete triggers for a specific 
    *        DMO interface. May be <code>null</code>.
    * @param assignTriggers
    *        Map of property names to sets of definitions which represent 
    *        assign triggers for a given property (for a specific DMO 
    *        interface). May be <code>null</code>.
    * @param noRecordAvailTriggers
    *        Set of definitions which represent "no record is available"  
    *        triggers for a specific DMO interface. May be <code>null</code>.
    */
   public DatabaseTriggerDelegate(Set<DatabaseTriggerDef> createTriggers,
                                  Set<DatabaseTriggerDef> deleteTriggers,
                                  Map<String, Set<DatabaseTriggerDef>> assignTriggers,
                                  Set<DatabaseTriggerDef> noRecordAvailTriggers)
   {
      this.createTriggers = createTriggers;
      this.deleteTriggers = deleteTriggers;
      this.assignTriggers = assignTriggers;
      this.noRecordAvailTriggers = noRecordAvailTriggers;
   }

   /**
    * Executes create triggers for the specific DMO.
    *
    * @param  dmo
    *         DMO which has triggered the event (i.e. it was just created).
    * @param  undo
    *         <code>true</code> if an undo is in effect.
    *         
    * @throws ConditionException
    *         May be thrown by trigger logic.
    */
   public void executeCreateTriggers(DataModelObject dmo, boolean undo)
   throws ConditionException
   {
      executeTriggers(createTriggers, undo, dmo, null);
   }
   
   /**
    * Executes delete triggers for the specific DMO.
    *
    * @param  dmo
    *         DMO which has triggered the event (i.e. it is about to be
    *         deleted). When the  <code>undo</code> flag is <code>false</code>, 
    *         this is the DMO proxy; else, it is the actual DMO record 
    *         instance.
    * @param  undo
    *         <code>true</code> if an undo is in effect.
    *
    * @throws ConditionException
    *         May be thrown by trigger logic.
    */
   public void executeDeleteTriggers(DataModelObject dmo, boolean undo)
   throws ConditionException
   {
      executeTriggers(deleteTriggers, undo, dmo, null);
   }

   /**
    * Executes assign triggers for the specific DMO.
    *
    * @param  dmo
    *         DMO which has triggered the event (i.e. a new value has been
    *         assigned to the specific field of this DMO). When the 
    *         <code>undo</code> flag is <code>false</code>, this is the DMO 
    *         proxy; else, it is the actual DMO record instance.
    * @param  propertyName
    *         Name of the field which was given a new value.
    * @param  oldFieldValue
    *         Old value of the field. 
    * @param  undo
    *         <code>true</code> if an undo is in effect.
    *
    * @throws ConditionException
    *         May be thrown by trigger logic.
    */
   public void executeAssignTriggers(DataModelObject dmo,
                                     String propertyName,
                                     BaseDataType oldFieldValue,
                                     boolean undo)
   throws ConditionException
   {
      if (assignTriggers != null)
      {
         Set<DatabaseTriggerDef> triggers = assignTriggers.get(propertyName);
         
         executeTriggers(triggers, undo, dmo, oldFieldValue);
      }
   }

   /**
    * Executes "no record is available" triggers for the specific record buffer
    * instance.
    *
    * @param  dmo
    *         DMO proxy which has triggered the event (i.e. the buffer was  
    *         expected to contain a record, but it does not).
    *
    * @throws ConditionException
    *         May be thrown by trigger logic.
    */
   public void executeNoRecordAvailableTriggers(DataModelObject dmo)
   throws ConditionException
   {
      executeTriggers(noRecordAvailTriggers, false, dmo, null);
   }
   
   /**
    * Execute the specified triggers.
    * 
    * @param   triggers
    *          The triggers to be executed.
    * @param   undo
    *          When <code>true</code>, the trigger must define a c'tor which
    *          takes a boolean value as a parameter, to save the undo state.
    * @param   dmo
    *          DMO which has triggered the event. When the <code>undo</code>
    *          flag is <code>false</code>, this is the DMO proxy; else, it
    *          is the actual DMO record instance.
    * @param   oldFieldValue
    *          Old value of the field; only for ASSIGN triggers.
    */
   private void executeTriggers(Set<DatabaseTriggerDef> triggers,
                                boolean undo,
                                DataModelObject dmo,
                                BaseDataType oldFieldValue)
   {
      if (triggers != null)
      {
         for (DatabaseTriggerDef triggerDef : triggers)
         {
            Class clazz = triggerDef.getTriggerClass();

            try
            {
               Object trigger = null;

               if (undo)
               {
                  if (triggerDef.isOnUndo())
                  {
                     trigger = clazz.getConstructor(boolean.class)
                                    .newInstance(new Object[] { true });
                  }
                  else
                  {
                     // this is a case of a trigger which doesn't need to 
                     // execute on undo
                     continue;
                  }
               }
               else
               {
                  trigger = clazz.newInstance();
               }

               // execute the trigger, depending on its type
               if (trigger instanceof CreateTrigger)
               {
                  ((CreateTrigger) trigger).execute(dmo);
               }
               else if (trigger instanceof DeleteTrigger)
               {
                  ((DeleteTrigger) trigger).execute(dmo);
               }
               else if (trigger instanceof AssignTrigger)
               {
                  ((AssignTrigger) trigger).execute(dmo, null, oldFieldValue);
               }
               else if (trigger instanceof NoRecordAvailableTrigger)
               {
                  ((NoRecordAvailableTrigger) trigger).execute(dmo);
               }
            }
            catch (IllegalAccessException e)
            {
               throw new RuntimeException(e);
            }
            catch (InstantiationException e)
            {
               throw new RuntimeException(e);
            } 
            catch (IllegalArgumentException e)
            {
               throw new RuntimeException(e);
            } 
            catch (SecurityException e)
            {
               throw new RuntimeException(e);
            } 
            catch (InvocationTargetException e)
            {
               throw new RuntimeException(e);
            } 
            catch (NoSuchMethodException e)
            {
               throw new RuntimeException(e);
            }  
         }
      }
   }
}
